searchData={"items":[{"type":"module","title":"Snakepit","doc":"Snakepit - A generalized high-performance pooler and session manager.\n\nExtracted from DSPex V3 pool implementation, Snakepit provides:\n- Concurrent worker initialization and management\n- Stateless pool system with session affinity \n- Generalized adapter pattern for any external process\n- High-performance OTP-based process management","ref":"Snakepit.html"},{"type":"module","title":"Basic Usage - Snakepit","doc":"# Configure in config/config.exs\n    config :snakepit,\n      pooling_enabled: true,\n      adapter_module: YourAdapter\n\n    # Execute commands on any available worker\n    {:ok, result} = Snakepit.execute(\"ping\", %{test: true})\n    \n    # Session-based execution with worker affinity\n    {:ok, result} = Snakepit.execute_in_session(\"my_session\", \"command\", %{})","ref":"Snakepit.html#module-basic-usage"},{"type":"module","title":"Domain-Specific Helpers - Snakepit","doc":"For ML/DSP workflows with program management, see `Snakepit.SessionHelpers`:\n\n    # ML program creation and execution\n    {:ok, result} = Snakepit.SessionHelpers.execute_program_command(\n      \"session_id\", \"create_program\", %{signature: \"input -> output\"}\n    )","ref":"Snakepit.html#module-domain-specific-helpers"},{"type":"function","title":"Snakepit.execute/3","doc":"Convenience function to execute commands on the pool.","ref":"Snakepit.html#execute/3"},{"type":"function","title":"Snakepit.execute_in_session/4","doc":"Executes a command in session context with worker affinity.\n\nThis function executes commands with session-based worker affinity,\nensuring that subsequent calls with the same session_id prefer\nthe same worker when possible for state continuity.\n\nArgs are passed through unchanged - no domain-specific enhancement.\nFor ML/DSP program workflows, use `Snakepit.SessionHelpers.execute_program_command/4`.","ref":"Snakepit.html#execute_in_session/4"},{"type":"function","title":"Snakepit.get_stats/1","doc":"Get pool statistics.","ref":"Snakepit.html#get_stats/1"},{"type":"function","title":"Snakepit.list_workers/1","doc":"List workers from the pool.","ref":"Snakepit.html#list_workers/1"},{"type":"behaviour","title":"Snakepit.Adapter","doc":"Behaviour for implementing adapters in Snakepit.\n\nAdapters define how to communicate with external processes (Python, Node.js, etc.)\nand what commands they support. This allows Snakepit to be truly generalized\nand support multiple ML frameworks or external systems.","ref":"Snakepit.Adapter.html"},{"type":"behaviour","title":"Required Callbacks - Snakepit.Adapter","doc":"- `executable_path/0` - Returns the path to the runtime executable (python3, node, etc.)\n- `script_path/0` - Returns the path to the external script to execute\n- `script_args/0` - Returns additional arguments for the script\n- `supported_commands/0` - Returns list of commands this adapter supports\n- `validate_command/2` - Validates a command and its arguments","ref":"Snakepit.Adapter.html#module-required-callbacks"},{"type":"behaviour","title":"Optional Callbacks - Snakepit.Adapter","doc":"- `process_response/2` - Post-process responses from the external process\n- `prepare_args/2` - Pre-process arguments before sending to external process","ref":"Snakepit.Adapter.html#module-optional-callbacks"},{"type":"behaviour","title":"Example Implementation - Snakepit.Adapter","doc":"defmodule MyApp.PythonMLAdapter do\n      @behaviour Snakepit.Adapter\n      \n      def executable_path, do: System.find_executable(\"python3\") || System.find_executable(\"python\")\n      def script_path, do: Path.join(:code.priv_dir(:my_app), \"python/ml_bridge.py\")\n      def script_args, do: [\"--mode\", \"pool-worker\"]\n      def supported_commands, do: [\"predict\", \"train\", \"ping\"]\n      \n      def validate_command(\"predict\", args) do\n        if Map.has_key?(args, :input), do: :ok, else: {:error, :missing_input}\n      end\n      def validate_command(\"ping\", _args), do: :ok\n      def validate_command(cmd, _), do: {:error, {:unsupported_command, cmd}}\n    end","ref":"Snakepit.Adapter.html#module-example-implementation"},{"type":"callback","title":"Snakepit.Adapter.command_timeout/2","doc":"Optional callback to get a command-specific timeout in milliseconds.\n\nThis allows adapters to specify appropriate timeouts for different\ncommands based on their expected execution time.","ref":"Snakepit.Adapter.html#c:command_timeout/2"},{"type":"function","title":"Snakepit.Adapter.default_command_timeout/2","doc":"Default implementation for command_timeout - returns 30 seconds.","ref":"Snakepit.Adapter.html#default_command_timeout/2"},{"type":"function","title":"Snakepit.Adapter.default_prepare_args/2","doc":"Default implementation for prepare_args - just returns args as-is.","ref":"Snakepit.Adapter.html#default_prepare_args/2"},{"type":"function","title":"Snakepit.Adapter.default_process_response/2","doc":"Default implementation for process_response - just returns the response as-is.","ref":"Snakepit.Adapter.html#default_process_response/2"},{"type":"callback","title":"Snakepit.Adapter.executable_path/0","doc":"Returns the path to the runtime executable.\n\nThis is the interpreter or runtime that will execute the script.\nExamples: \"python3\", \"node\", \"ruby\", \"R\", etc.","ref":"Snakepit.Adapter.html#c:executable_path/0"},{"type":"callback","title":"Snakepit.Adapter.prepare_args/2","doc":"Optional callback to prepare arguments before sending to external process.\n\nThis allows adapters to transform arguments into the format\nexpected by their external script.","ref":"Snakepit.Adapter.html#c:prepare_args/2"},{"type":"callback","title":"Snakepit.Adapter.process_response/2","doc":"Optional callback to process responses from the external process.\n\nThis allows adapters to transform or validate responses before\nthey're returned to the caller.","ref":"Snakepit.Adapter.html#c:process_response/2"},{"type":"callback","title":"Snakepit.Adapter.script_args/0","doc":"Returns additional command-line arguments for the script.\n\nThese arguments will be passed to the script when it's started.\nCommon examples: [\"--mode\", \"pool-worker\"], [\"--config\", \"prod\"]","ref":"Snakepit.Adapter.html#c:script_args/0"},{"type":"callback","title":"Snakepit.Adapter.script_path/0","doc":"Returns the path to the external script that will be executed.\n\nThis should be an absolute path to a script that implements the\nbridge protocol for communication with Snakepit.","ref":"Snakepit.Adapter.html#c:script_path/0"},{"type":"callback","title":"Snakepit.Adapter.supported_commands/0","doc":"Returns a list of commands that this adapter supports.\n\nThis is used for validation and documentation purposes.","ref":"Snakepit.Adapter.html#c:supported_commands/0"},{"type":"callback","title":"Snakepit.Adapter.validate_command/2","doc":"Validates that a command and its arguments are valid for this adapter.\n\nReturns `:ok` if valid, `{:error, reason}` if invalid.","ref":"Snakepit.Adapter.html#c:validate_command/2"},{"type":"module","title":"Snakepit.Adapters.GenericJavaScript","doc":"Generic JavaScript/Node.js adapter for Snakepit.\n\nThis adapter provides a simple, framework-agnostic bridge to Node.js\nwithout any external dependencies. It's useful for:\n\n- Testing the Snakepit infrastructure with JavaScript\n- Simple computational tasks in Node.js\n- Web scraping or API calls using JavaScript\n- As a template for building JavaScript-based adapters","ref":"Snakepit.Adapters.GenericJavaScript.html"},{"type":"module","title":"Supported Commands - Snakepit.Adapters.GenericJavaScript","doc":"- `ping` - Health check and basic info\n- `echo` - Echo arguments back (useful for testing)\n- `compute` - Simple math operations (add, subtract, multiply, divide)\n- `info` - Bridge and system information\n- `random` - Generate random numbers with various distributions","ref":"Snakepit.Adapters.GenericJavaScript.html#module-supported-commands"},{"type":"module","title":"Configuration - Snakepit.Adapters.GenericJavaScript","doc":"config :snakepit,\n      adapter_module: Snakepit.Adapters.GenericJavaScript","ref":"Snakepit.Adapters.GenericJavaScript.html#module-configuration"},{"type":"module","title":"Usage Examples - Snakepit.Adapters.GenericJavaScript","doc":"# Health check\n    {:ok, result} = Snakepit.execute(\"ping\", %{test: true})\n    \n    # Echo test\n    {:ok, result} = Snakepit.execute(\"echo\", %{message: \"hello\"})\n    \n    # Simple computation\n    {:ok, result} = Snakepit.execute(\"compute\", %{\n      operation: \"add\",\n      a: 5, \n      b: 3\n    })\n    \n    # Random number generation\n    {:ok, result} = Snakepit.execute(\"random\", %{\n      type: \"uniform\",\n      min: 1,\n      max: 100\n    })\n    \n    # Get bridge info\n    {:ok, result} = Snakepit.execute(\"info\", %{})","ref":"Snakepit.Adapters.GenericJavaScript.html#module-usage-examples"},{"type":"module","title":"Snakepit.Adapters.GenericPython","doc":"Generic Python adapter for Snakepit.\n\nThis adapter provides a simple, framework-agnostic bridge to Python\nwithout any external ML dependencies. It's useful for:\n\n- Testing the Snakepit infrastructure\n- Simple computational tasks\n- As a template for building your own adapters","ref":"Snakepit.Adapters.GenericPython.html"},{"type":"module","title":"Supported Commands - Snakepit.Adapters.GenericPython","doc":"- `ping` - Health check and basic info\n- `echo` - Echo arguments back (useful for testing)\n- `compute` - Simple math operations (add, subtract, multiply, divide)\n- `info` - Bridge and system information","ref":"Snakepit.Adapters.GenericPython.html#module-supported-commands"},{"type":"module","title":"Configuration - Snakepit.Adapters.GenericPython","doc":"config :snakepit,\n      adapter_module: Snakepit.Adapters.GenericPython","ref":"Snakepit.Adapters.GenericPython.html#module-configuration"},{"type":"module","title":"Usage Examples - Snakepit.Adapters.GenericPython","doc":"# Health check\n    {:ok, result} = Snakepit.execute(\"ping\", %{test: true})\n    \n    # Echo test\n    {:ok, result} = Snakepit.execute(\"echo\", %{message: \"hello\"})\n    \n    # Simple computation\n    {:ok, result} = Snakepit.execute(\"compute\", %{\n      operation: \"add\",\n      a: 5, \n      b: 3\n    })\n    \n    # Get bridge info\n    {:ok, result} = Snakepit.execute(\"info\", %{})","ref":"Snakepit.Adapters.GenericPython.html#module-usage-examples"},{"type":"module","title":"Snakepit.Application","doc":"Application supervisor for Snakepit pooler.\n\nStarts the core infrastructure:\n- Registry for worker process registration\n- StarterRegistry for worker starter supervisors\n- ProcessRegistry for external PID tracking\n- SessionStore for session management\n- WorkerSupervisor for managing worker processes\n- Pool manager for request distribution","ref":"Snakepit.Application.html"},{"type":"module","title":"Snakepit.Bridge.Protocol","doc":"Wire protocol for Python bridge communication.\n\nThis module handles the encoding and decoding of messages between Elixir and Python\nprocesses using a JSON-based protocol with 4-byte length headers for proper framing.","ref":"Snakepit.Bridge.Protocol.html"},{"type":"module","title":"Protocol Format - Snakepit.Bridge.Protocol","doc":"Each message consists of:\n1. 4-byte big-endian length header indicating the JSON payload size\n2. JSON payload containing the actual message data","ref":"Snakepit.Bridge.Protocol.html#module-protocol-format"},{"type":"module","title":"Request Format - Snakepit.Bridge.Protocol","doc":"%{\n      \"id\" => integer(),\n      \"command\" => string(),\n      \"args\" => map(),\n      \"timestamp\" => iso8601_string()\n    }","ref":"Snakepit.Bridge.Protocol.html#module-request-format"},{"type":"module","title":"Response Format - Snakepit.Bridge.Protocol","doc":"# Success\n    %{\n      \"id\" => integer(),\n      \"success\" => true,\n      \"result\" => any(),\n      \"timestamp\" => iso8601_string()\n    }\n\n    # Error\n    %{\n      \"id\" => integer(),\n      \"success\" => false,\n      \"error\" => string(),\n      \"timestamp\" => iso8601_string()\n    }","ref":"Snakepit.Bridge.Protocol.html#module-response-format"},{"type":"module","title":"Features - Snakepit.Bridge.Protocol","doc":"- Request/response correlation using unique IDs\n- Timestamping for debugging and metrics\n- Structured error handling\n- JSON encoding for cross-language compatibility\n- Length-prefixed framing for reliable message boundaries","ref":"Snakepit.Bridge.Protocol.html#module-features"},{"type":"type","title":"Snakepit.Bridge.Protocol.args/0","doc":"","ref":"Snakepit.Bridge.Protocol.html#t:args/0"},{"type":"type","title":"Snakepit.Bridge.Protocol.command/0","doc":"","ref":"Snakepit.Bridge.Protocol.html#t:command/0"},{"type":"function","title":"Snakepit.Bridge.Protocol.create_error_response/2","doc":"Creates a standardized error response.\n\nGenerates a properly formatted error response that can be sent back\nto the caller when an error occurs during request processing.","ref":"Snakepit.Bridge.Protocol.html#create_error_response/2"},{"type":"function","title":"Examples - Snakepit.Bridge.Protocol.create_error_response/2","doc":"iex> Snakepit.Bridge.Protocol.create_error_response(1, \"Command failed\")\n    %{\"id\" => 1, \"success\" => false, \"error\" => \"Command failed\", \"timestamp\" => _}","ref":"Snakepit.Bridge.Protocol.html#create_error_response/2-examples"},{"type":"function","title":"Snakepit.Bridge.Protocol.create_success_response/2","doc":"Creates a standardized success response.\n\nGenerates a properly formatted success response that can be sent back\nto the caller when a request completes successfully.","ref":"Snakepit.Bridge.Protocol.html#create_success_response/2"},{"type":"function","title":"Examples - Snakepit.Bridge.Protocol.create_success_response/2","doc":"iex> Snakepit.Bridge.Protocol.create_success_response(1, %{\"status\" => \"ok\"})\n    %{\"id\" => 1, \"success\" => true, \"result\" => %{\"status\" => \"ok\"}, \"timestamp\" => _}","ref":"Snakepit.Bridge.Protocol.html#create_success_response/2-examples"},{"type":"function","title":"Snakepit.Bridge.Protocol.decode_response/1","doc":"Decodes a response received from the Python bridge.\n\nParses JSON response data and extracts the request ID, success status,\nand either the result or error information.","ref":"Snakepit.Bridge.Protocol.html#decode_response/1"},{"type":"function","title":"Examples - Snakepit.Bridge.Protocol.decode_response/1","doc":"iex> json = ~s({\"id\":1,\"success\":true,\"result\":{\"status\":\"ok\"}})\n    iex> Snakepit.Bridge.Protocol.decode_response(json)\n    {:ok, 1, %{\"status\" => \"ok\"}}\n\n    iex> json = ~s({\"id\":2,\"success\":false,\"error\":\"Something went wrong\"})\n    iex> Snakepit.Bridge.Protocol.decode_response(json)\n    {:error, 2, \"Something went wrong\"}\n\n    iex> Snakepit.Bridge.Protocol.decode_response(\"invalid json\")\n    {:error, :decode_error}","ref":"Snakepit.Bridge.Protocol.html#decode_response/1-examples"},{"type":"function","title":"Snakepit.Bridge.Protocol.encode_request/3","doc":"Encodes a request for sending to the Python bridge.\n\nCreates a properly formatted request message with a unique ID, command,\narguments, and timestamp. The message is encoded as JSON for transmission.","ref":"Snakepit.Bridge.Protocol.html#encode_request/3"},{"type":"function","title":"Examples - Snakepit.Bridge.Protocol.encode_request/3","doc":"iex> Snakepit.Bridge.Protocol.encode_request(1, :ping, %{})\n    ~s({\"id\":1,\"command\":\"ping\",\"args\":{},\"timestamp\":\"2024-01-01T00:00:00Z\"})\n\n    iex> Snakepit.Bridge.Protocol.encode_request(2, \"create_program\", %{signature: %{}})\n    ~s({\"id\":2,\"command\":\"create_program\",\"args\":{\"signature\":{}},\"timestamp\":\"2024-01-01T00:00:00Z\"})","ref":"Snakepit.Bridge.Protocol.html#encode_request/3-examples"},{"type":"type","title":"Snakepit.Bridge.Protocol.error_reason/0","doc":"","ref":"Snakepit.Bridge.Protocol.html#t:error_reason/0"},{"type":"function","title":"Snakepit.Bridge.Protocol.extract_request_id/1","doc":"Extracts the request ID from a message.\n\nSafely extracts the request ID from either a request or response message.\nReturns nil if the ID is not present or not valid.","ref":"Snakepit.Bridge.Protocol.html#extract_request_id/1"},{"type":"function","title":"Examples - Snakepit.Bridge.Protocol.extract_request_id/1","doc":"iex> Snakepit.Bridge.Protocol.extract_request_id(%{\"id\" => 42})\n    42\n\n    iex> Snakepit.Bridge.Protocol.extract_request_id(%{\"command\" => \"ping\"})\n    nil","ref":"Snakepit.Bridge.Protocol.html#extract_request_id/1-examples"},{"type":"function","title":"Snakepit.Bridge.Protocol.generate_request_id/0","doc":"Generates a unique request ID.\n\nCreates a monotonically increasing request ID that can be used for\nrequest/response correlation. Uses System.unique_integer for \nhigh performance without serialization bottlenecks.","ref":"Snakepit.Bridge.Protocol.html#generate_request_id/0"},{"type":"type","title":"Snakepit.Bridge.Protocol.request_id/0","doc":"","ref":"Snakepit.Bridge.Protocol.html#t:request_id/0"},{"type":"type","title":"Snakepit.Bridge.Protocol.result/0","doc":"","ref":"Snakepit.Bridge.Protocol.html#t:result/0"},{"type":"function","title":"Snakepit.Bridge.Protocol.validate_request/1","doc":"Validates that a request has all required fields.\n\nChecks that a request map contains the necessary fields and that they\nhave the correct types.","ref":"Snakepit.Bridge.Protocol.html#validate_request/1"},{"type":"function","title":"Examples - Snakepit.Bridge.Protocol.validate_request/1","doc":"iex> request = %{\"id\" => 1, \"command\" => \"ping\", \"args\" => %{}}\n    iex> Snakepit.Bridge.Protocol.validate_request(request)\n    :ok\n\n    iex> Snakepit.Bridge.Protocol.validate_request(%{\"id\" => 1})\n    {:error, \"Missing required field: command\"}","ref":"Snakepit.Bridge.Protocol.html#validate_request/1-examples"},{"type":"function","title":"Snakepit.Bridge.Protocol.validate_response/1","doc":"Validates that a response has all required fields.\n\nChecks that a response map contains the necessary fields for either\na success or error response.","ref":"Snakepit.Bridge.Protocol.html#validate_response/1"},{"type":"function","title":"Examples - Snakepit.Bridge.Protocol.validate_response/1","doc":"iex> response = %{\"id\" => 1, \"success\" => true, \"result\" => %{}}\n    iex> Snakepit.Bridge.Protocol.validate_response(response)\n    :ok\n\n    iex> response = %{\"id\" => 1, \"success\" => false, \"error\" => \"Failed\"}\n    iex> Snakepit.Bridge.Protocol.validate_response(response)\n    :ok\n\n    iex> Snakepit.Bridge.Protocol.validate_response(%{\"id\" => 1})\n    {:error, \"Missing required field: success\"}","ref":"Snakepit.Bridge.Protocol.html#validate_response/1-examples"},{"type":"module","title":"Snakepit.Bridge.Session","doc":"Session data structure for centralized session management.\n\nThis struct represents a session in the centralized session store,\ncontaining all session-related data including programs, metadata,\nand lifecycle information.","ref":"Snakepit.Bridge.Session.html"},{"type":"function","title":"Snakepit.Bridge.Session.delete_program/2","doc":"Removes a program from the session.","ref":"Snakepit.Bridge.Session.html#delete_program/2"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.delete_program/2","doc":"- `session` - The session to update\n- `program_id` - The program identifier to remove","ref":"Snakepit.Bridge.Session.html#delete_program/2-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.Session.delete_program/2","doc":"Updated session with the program removed.","ref":"Snakepit.Bridge.Session.html#delete_program/2-returns"},{"type":"function","title":"Snakepit.Bridge.Session.expired?/2","doc":"Checks if a session has expired based on its TTL.","ref":"Snakepit.Bridge.Session.html#expired?/2"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.expired?/2","doc":"- `session` - The session to check\n- `current_time` - Optional current time (defaults to current monotonic time)","ref":"Snakepit.Bridge.Session.html#expired?/2-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.Session.expired?/2","doc":"`true` if the session has expired, `false` otherwise.","ref":"Snakepit.Bridge.Session.html#expired?/2-returns"},{"type":"function","title":"Snakepit.Bridge.Session.get_metadata/3","doc":"Gets metadata from the session.","ref":"Snakepit.Bridge.Session.html#get_metadata/3"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.get_metadata/3","doc":"- `session` - The session to query\n- `key` - The metadata key\n- `default` - Default value if key not found","ref":"Snakepit.Bridge.Session.html#get_metadata/3-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.Session.get_metadata/3","doc":"The metadata value or the default.","ref":"Snakepit.Bridge.Session.html#get_metadata/3-returns"},{"type":"function","title":"Snakepit.Bridge.Session.get_program/2","doc":"Gets a program from the session.","ref":"Snakepit.Bridge.Session.html#get_program/2"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.get_program/2","doc":"- `session` - The session to query\n- `program_id` - The program identifier","ref":"Snakepit.Bridge.Session.html#get_program/2-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.Session.get_program/2","doc":"`{:ok, program_data}` if found, `{:error, :not_found}` if not found.","ref":"Snakepit.Bridge.Session.html#get_program/2-returns"},{"type":"function","title":"Snakepit.Bridge.Session.new/2","doc":"Creates a new session with the given ID and options.","ref":"Snakepit.Bridge.Session.html#new/2"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.new/2","doc":"- `id` - Unique session identifier\n- `opts` - Keyword list of options:\n  - `:ttl` - Time-to-live in seconds (default: 3600)\n  - `:metadata` - Initial metadata map (default: %{})\n  - `:programs` - Initial programs map (default: %{})","ref":"Snakepit.Bridge.Session.html#new/2-parameters"},{"type":"function","title":"Examples - Snakepit.Bridge.Session.new/2","doc":"iex> Session.new(\"session_123\")\n    %Session{id: \"session_123\", programs: %{}, metadata: %{}, ...}\n    \n    iex> Session.new(\"session_456\", ttl: 7200, metadata: %{user_id: \"user_1\"})\n    %Session{id: \"session_456\", ttl: 7200, metadata: %{user_id: \"user_1\"}, ...}","ref":"Snakepit.Bridge.Session.html#new/2-examples"},{"type":"function","title":"Snakepit.Bridge.Session.put_metadata/3","doc":"Updates session metadata.","ref":"Snakepit.Bridge.Session.html#put_metadata/3"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.put_metadata/3","doc":"- `session` - The session to update\n- `key` - The metadata key\n- `value` - The metadata value","ref":"Snakepit.Bridge.Session.html#put_metadata/3-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.Session.put_metadata/3","doc":"Updated session with the metadata updated.","ref":"Snakepit.Bridge.Session.html#put_metadata/3-returns"},{"type":"function","title":"Snakepit.Bridge.Session.put_program/3","doc":"Adds or updates a program in the session.","ref":"Snakepit.Bridge.Session.html#put_program/3"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.put_program/3","doc":"- `session` - The session to update\n- `program_id` - The program identifier\n- `program_data` - The program data to store","ref":"Snakepit.Bridge.Session.html#put_program/3-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.Session.put_program/3","doc":"Updated session with the program added/updated.","ref":"Snakepit.Bridge.Session.html#put_program/3-returns"},{"type":"type","title":"Snakepit.Bridge.Session.t/0","doc":"","ref":"Snakepit.Bridge.Session.html#t:t/0"},{"type":"function","title":"Snakepit.Bridge.Session.touch/1","doc":"Updates the last_accessed timestamp to the current time.","ref":"Snakepit.Bridge.Session.html#touch/1"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.touch/1","doc":"- `session` - The session to touch","ref":"Snakepit.Bridge.Session.html#touch/1-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.Session.touch/1","doc":"Updated session with current last_accessed timestamp.","ref":"Snakepit.Bridge.Session.html#touch/1-returns"},{"type":"function","title":"Snakepit.Bridge.Session.validate/1","doc":"Validates that a session struct has all required fields and valid data.","ref":"Snakepit.Bridge.Session.html#validate/1"},{"type":"function","title":"Parameters - Snakepit.Bridge.Session.validate/1","doc":"- `session` - The session to validate","ref":"Snakepit.Bridge.Session.html#validate/1-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.Session.validate/1","doc":"`:ok` if valid, `{:error, reason}` if invalid.","ref":"Snakepit.Bridge.Session.html#validate/1-returns"},{"type":"module","title":"Snakepit.Bridge.SessionStore","doc":"Centralized session store using ETS for high-performance session management.\n\nThis GenServer manages a centralized ETS table for storing session data,\nproviding CRUD operations, TTL-based expiration, and automatic cleanup.\nThe store is designed for high concurrency with optimized ETS settings.","ref":"Snakepit.Bridge.SessionStore.html"},{"type":"function","title":"Snakepit.Bridge.SessionStore.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Snakepit.Bridge.SessionStore.html#child_spec/1"},{"type":"function","title":"Snakepit.Bridge.SessionStore.cleanup_expired_sessions/0","doc":"Manually triggers cleanup of expired sessions.","ref":"Snakepit.Bridge.SessionStore.html#cleanup_expired_sessions/0"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.cleanup_expired_sessions/0","doc":"The number of sessions that were cleaned up.","ref":"Snakepit.Bridge.SessionStore.html#cleanup_expired_sessions/0-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.cleanup_expired_sessions/1","doc":"","ref":"Snakepit.Bridge.SessionStore.html#cleanup_expired_sessions/1"},{"type":"function","title":"Snakepit.Bridge.SessionStore.create_session/2","doc":"Creates a new session with the given ID and options.","ref":"Snakepit.Bridge.SessionStore.html#create_session/2"},{"type":"function","title":"Parameters - Snakepit.Bridge.SessionStore.create_session/2","doc":"- `session_id` - Unique session identifier\n- `opts` - Keyword list of options passed to Session.new/2","ref":"Snakepit.Bridge.SessionStore.html#create_session/2-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.create_session/2","doc":"`{:ok, session}` if successful, `{:error, reason}` if failed.","ref":"Snakepit.Bridge.SessionStore.html#create_session/2-returns"},{"type":"function","title":"Examples - Snakepit.Bridge.SessionStore.create_session/2","doc":"{:ok, session} = SessionStore.create_session(\"session_123\")\n    {:ok, session} = SessionStore.create_session(\"session_456\", ttl: 7200)","ref":"Snakepit.Bridge.SessionStore.html#create_session/2-examples"},{"type":"function","title":"Snakepit.Bridge.SessionStore.create_session/3","doc":"","ref":"Snakepit.Bridge.SessionStore.html#create_session/3"},{"type":"function","title":"Snakepit.Bridge.SessionStore.delete_global_program/1","doc":"Deletes a globally stored program.","ref":"Snakepit.Bridge.SessionStore.html#delete_global_program/1"},{"type":"function","title":"Parameters - Snakepit.Bridge.SessionStore.delete_global_program/1","doc":"- `program_id` - The program identifier","ref":"Snakepit.Bridge.SessionStore.html#delete_global_program/1-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.delete_global_program/1","doc":"`:ok` always (idempotent operation).","ref":"Snakepit.Bridge.SessionStore.html#delete_global_program/1-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.delete_global_program/2","doc":"","ref":"Snakepit.Bridge.SessionStore.html#delete_global_program/2"},{"type":"function","title":"Snakepit.Bridge.SessionStore.delete_session/1","doc":"Deletes a session by ID.","ref":"Snakepit.Bridge.SessionStore.html#delete_session/1"},{"type":"function","title":"Parameters - Snakepit.Bridge.SessionStore.delete_session/1","doc":"- `session_id` - The session identifier","ref":"Snakepit.Bridge.SessionStore.html#delete_session/1-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.delete_session/1","doc":"`:ok` always (idempotent operation).","ref":"Snakepit.Bridge.SessionStore.html#delete_session/1-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.delete_session/2","doc":"","ref":"Snakepit.Bridge.SessionStore.html#delete_session/2"},{"type":"function","title":"Snakepit.Bridge.SessionStore.get_global_program/1","doc":"Retrieves a globally stored program.","ref":"Snakepit.Bridge.SessionStore.html#get_global_program/1"},{"type":"function","title":"Parameters - Snakepit.Bridge.SessionStore.get_global_program/1","doc":"- `program_id` - The program identifier","ref":"Snakepit.Bridge.SessionStore.html#get_global_program/1-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.get_global_program/1","doc":"`{:ok, program_data}` if found, `{:error, :not_found}` if not found.","ref":"Snakepit.Bridge.SessionStore.html#get_global_program/1-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.get_global_program/2","doc":"","ref":"Snakepit.Bridge.SessionStore.html#get_global_program/2"},{"type":"function","title":"Snakepit.Bridge.SessionStore.get_program/2","doc":"Gets a program from a session.","ref":"Snakepit.Bridge.SessionStore.html#get_program/2"},{"type":"function","title":"Snakepit.Bridge.SessionStore.get_session/1","doc":"Gets a session by ID, automatically updating the last_accessed timestamp.","ref":"Snakepit.Bridge.SessionStore.html#get_session/1"},{"type":"function","title":"Parameters - Snakepit.Bridge.SessionStore.get_session/1","doc":"- `session_id` - The session identifier","ref":"Snakepit.Bridge.SessionStore.html#get_session/1-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.get_session/1","doc":"`{:ok, session}` if found, `{:error, :not_found}` if not found.","ref":"Snakepit.Bridge.SessionStore.html#get_session/1-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.get_session/2","doc":"","ref":"Snakepit.Bridge.SessionStore.html#get_session/2"},{"type":"function","title":"Snakepit.Bridge.SessionStore.get_stats/0","doc":"Gets statistics about the session store.","ref":"Snakepit.Bridge.SessionStore.html#get_stats/0"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.get_stats/0","doc":"A map containing various statistics about the session store.","ref":"Snakepit.Bridge.SessionStore.html#get_stats/0-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.get_stats/1","doc":"","ref":"Snakepit.Bridge.SessionStore.html#get_stats/1"},{"type":"function","title":"Snakepit.Bridge.SessionStore.list_sessions/0","doc":"Lists all active session IDs.","ref":"Snakepit.Bridge.SessionStore.html#list_sessions/0"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.list_sessions/0","doc":"A list of all active session IDs.","ref":"Snakepit.Bridge.SessionStore.html#list_sessions/0-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.list_sessions/1","doc":"","ref":"Snakepit.Bridge.SessionStore.html#list_sessions/1"},{"type":"function","title":"Snakepit.Bridge.SessionStore.session_exists?/1","doc":"Checks if a session exists.","ref":"Snakepit.Bridge.SessionStore.html#session_exists?/1"},{"type":"function","title":"Parameters - Snakepit.Bridge.SessionStore.session_exists?/1","doc":"- `session_id` - The session identifier","ref":"Snakepit.Bridge.SessionStore.html#session_exists?/1-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.session_exists?/1","doc":"`true` if the session exists, `false` otherwise.","ref":"Snakepit.Bridge.SessionStore.html#session_exists?/1-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.session_exists?/2","doc":"","ref":"Snakepit.Bridge.SessionStore.html#session_exists?/2"},{"type":"function","title":"Snakepit.Bridge.SessionStore.start_link/1","doc":"Starts the SessionStore GenServer.","ref":"Snakepit.Bridge.SessionStore.html#start_link/1"},{"type":"function","title":"Options - Snakepit.Bridge.SessionStore.start_link/1","doc":"- `:name` - The name to register the GenServer (default: __MODULE__)\n- `:table_name` - The ETS table name (default: :snakepit_sessions)\n- `:cleanup_interval` - Cleanup interval in milliseconds (default: 60_000)\n- `:default_ttl` - Default TTL for sessions in seconds (default: 3600)","ref":"Snakepit.Bridge.SessionStore.html#start_link/1-options"},{"type":"function","title":"Snakepit.Bridge.SessionStore.store_global_program/2","doc":"Stores a program globally, accessible to any worker.\n\nThis is used for anonymous operations where programs need to be\naccessible across different pool workers.","ref":"Snakepit.Bridge.SessionStore.html#store_global_program/2"},{"type":"function","title":"Parameters - Snakepit.Bridge.SessionStore.store_global_program/2","doc":"- `program_id` - Unique program identifier\n- `program_data` - Program data to store","ref":"Snakepit.Bridge.SessionStore.html#store_global_program/2-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.store_global_program/2","doc":"`:ok` if successful, `{:error, reason}` if failed.","ref":"Snakepit.Bridge.SessionStore.html#store_global_program/2-returns"},{"type":"function","title":"Snakepit.Bridge.SessionStore.store_global_program/3","doc":"","ref":"Snakepit.Bridge.SessionStore.html#store_global_program/3"},{"type":"function","title":"Snakepit.Bridge.SessionStore.store_program/3","doc":"Stores a program in a session.","ref":"Snakepit.Bridge.SessionStore.html#store_program/3"},{"type":"function","title":"Snakepit.Bridge.SessionStore.store_worker_session/2","doc":"Stores worker-session affinity mapping.","ref":"Snakepit.Bridge.SessionStore.html#store_worker_session/2"},{"type":"function","title":"Snakepit.Bridge.SessionStore.update_program/3","doc":"Updates a program in a session.","ref":"Snakepit.Bridge.SessionStore.html#update_program/3"},{"type":"function","title":"Snakepit.Bridge.SessionStore.update_session/2","doc":"Updates a session using the provided update function.\n\nThe update function receives the current session and should return\nthe updated session. The operation is atomic.","ref":"Snakepit.Bridge.SessionStore.html#update_session/2"},{"type":"function","title":"Parameters - Snakepit.Bridge.SessionStore.update_session/2","doc":"- `session_id` - The session identifier\n- `update_fn` - Function that takes a session and returns an updated session","ref":"Snakepit.Bridge.SessionStore.html#update_session/2-parameters"},{"type":"function","title":"Returns - Snakepit.Bridge.SessionStore.update_session/2","doc":"`{:ok, updated_session}` if successful, `{:error, reason}` if failed.","ref":"Snakepit.Bridge.SessionStore.html#update_session/2-returns"},{"type":"function","title":"Examples - Snakepit.Bridge.SessionStore.update_session/2","doc":"{:ok, session} = SessionStore.update_session(\"session_123\", fn session ->\n      Session.put_program(session, \"prog_1\", %{data: \"example\"})\n    end)","ref":"Snakepit.Bridge.SessionStore.html#update_session/2-examples"},{"type":"function","title":"Snakepit.Bridge.SessionStore.update_session/3","doc":"","ref":"Snakepit.Bridge.SessionStore.html#update_session/3"},{"type":"module","title":"Snakepit.Pool","doc":"Pool manager for external workers with concurrent initialization.\n\nFeatures:\n- Concurrent worker startup (all workers start in parallel)\n- Simple queue-based request distribution\n- Non-blocking async execution\n- Automatic request queueing when workers are busy\n- Adapter-based support for any external process","ref":"Snakepit.Pool.html"},{"type":"function","title":"Snakepit.Pool.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Snakepit.Pool.html#child_spec/1"},{"type":"function","title":"Snakepit.Pool.execute/3","doc":"Executes a command on any available worker.","ref":"Snakepit.Pool.html#execute/3"},{"type":"function","title":"Snakepit.Pool.get_stats/1","doc":"Gets pool statistics.","ref":"Snakepit.Pool.html#get_stats/1"},{"type":"function","title":"Snakepit.Pool.list_workers/1","doc":"Lists all worker IDs in the pool.","ref":"Snakepit.Pool.html#list_workers/1"},{"type":"function","title":"Snakepit.Pool.start_link/1","doc":"Starts the pool manager.","ref":"Snakepit.Pool.html#start_link/1"},{"type":"module","title":"Snakepit.Pool.ApplicationCleanup","doc":"Provides hard guarantees for worker process cleanup when the application exits.\n\nThis module ensures that NO worker processes survive application shutdown,\npreventing orphaned processes while still allowing normal pool operations.","ref":"Snakepit.Pool.ApplicationCleanup.html"},{"type":"function","title":"Snakepit.Pool.ApplicationCleanup.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Snakepit.Pool.ApplicationCleanup.html#child_spec/1"},{"type":"function","title":"Snakepit.Pool.ApplicationCleanup.force_cleanup_all/0","doc":"Force cleanup all tracked worker processes.","ref":"Snakepit.Pool.ApplicationCleanup.html#force_cleanup_all/0"},{"type":"function","title":"Snakepit.Pool.ApplicationCleanup.init/1","doc":"Callback implementation for  GenServer.init/1 .","ref":"Snakepit.Pool.ApplicationCleanup.html#init/1"},{"type":"function","title":"Snakepit.Pool.ApplicationCleanup.start_link/1","doc":"","ref":"Snakepit.Pool.ApplicationCleanup.html#start_link/1"},{"type":"module","title":"Snakepit.Pool.ProcessRegistry","doc":"Registry for tracking external worker processes with OS-level PID management.\n\nThis module maintains a mapping between:\n- Worker IDs\n- Elixir worker PIDs\n- External process PIDs\n- Process fingerprints\n\nEnables robust orphaned process detection and cleanup.","ref":"Snakepit.Pool.ProcessRegistry.html"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Snakepit.Pool.ProcessRegistry.html#child_spec/1"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.cleanup_dead_workers/0","doc":"Cleans up dead worker entries from the registry.","ref":"Snakepit.Pool.ProcessRegistry.html#cleanup_dead_workers/0"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.get_active_process_pids/0","doc":"Gets all active external process PIDs from registered workers.","ref":"Snakepit.Pool.ProcessRegistry.html#get_active_process_pids/0"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.get_all_process_pids/0","doc":"Gets all registered external process PIDs, regardless of worker status.\n\nThis is useful during shutdown when workers may have been terminated\nbut external processes still need cleanup.","ref":"Snakepit.Pool.ProcessRegistry.html#get_all_process_pids/0"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.get_stats/0","doc":"Gets registry statistics.","ref":"Snakepit.Pool.ProcessRegistry.html#get_stats/0"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.get_worker_info/1","doc":"Gets information for a specific worker.","ref":"Snakepit.Pool.ProcessRegistry.html#get_worker_info/1"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.get_workers_by_fingerprint/1","doc":"Gets workers with specific fingerprints.","ref":"Snakepit.Pool.ProcessRegistry.html#get_workers_by_fingerprint/1"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.list_all_workers/0","doc":"Gets all registered worker information.","ref":"Snakepit.Pool.ProcessRegistry.html#list_all_workers/0"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.register_worker/4","doc":"Registers a worker with its external process information.","ref":"Snakepit.Pool.ProcessRegistry.html#register_worker/4"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.start_link/1","doc":"","ref":"Snakepit.Pool.ProcessRegistry.html#start_link/1"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.unregister_worker/1","doc":"Unregisters a worker from tracking.","ref":"Snakepit.Pool.ProcessRegistry.html#unregister_worker/1"},{"type":"function","title":"Snakepit.Pool.ProcessRegistry.validate_workers/0","doc":"Validates that all registered workers are still alive.\nReturns a list of dead workers that should be cleaned up.","ref":"Snakepit.Pool.ProcessRegistry.html#validate_workers/0"},{"type":"module","title":"Snakepit.Pool.Registry","doc":"Registry for pool worker processes.\n\nThis is a thin wrapper around Elixir's Registry that provides:\n- Consistent naming for worker processes\n- Easy migration path to distributed registry (Horde)\n- Helper functions for worker lookup","ref":"Snakepit.Pool.Registry.html"},{"type":"function","title":"Snakepit.Pool.Registry.child_spec/1","doc":"Returns the child spec for the registry.","ref":"Snakepit.Pool.Registry.html#child_spec/1"},{"type":"function","title":"Snakepit.Pool.Registry.get_worker_id_by_pid/1","doc":"Get worker_id from PID for O(1) lookups in :DOWN messages.","ref":"Snakepit.Pool.Registry.html#get_worker_id_by_pid/1"},{"type":"function","title":"Snakepit.Pool.Registry.get_worker_pid/1","doc":"Gets the PID for a worker ID.","ref":"Snakepit.Pool.Registry.html#get_worker_pid/1"},{"type":"function","title":"Snakepit.Pool.Registry.list_workers/0","doc":"Lists all registered worker IDs.","ref":"Snakepit.Pool.Registry.html#list_workers/0"},{"type":"function","title":"Snakepit.Pool.Registry.register_worker/2","doc":"Register a worker with metadata for O(1) reverse lookups.\nThis is only used for manual registration - workers started with via_tuple are already registered.","ref":"Snakepit.Pool.Registry.html#register_worker/2"},{"type":"function","title":"Snakepit.Pool.Registry.via_tuple/1","doc":"Returns a via tuple for registering/looking up a worker process.","ref":"Snakepit.Pool.Registry.html#via_tuple/1"},{"type":"function","title":"Examples - Snakepit.Pool.Registry.via_tuple/1","doc":"iex> Snakepit.Pool.Registry.via_tuple(\"worker_123\")\n    {:via, Registry, {Snakepit.Pool.Registry, \"worker_123\"}}","ref":"Snakepit.Pool.Registry.html#via_tuple/1-examples"},{"type":"function","title":"Snakepit.Pool.Registry.worker_count/0","doc":"Counts the number of registered workers.","ref":"Snakepit.Pool.Registry.html#worker_count/0"},{"type":"function","title":"Snakepit.Pool.Registry.worker_exists?/1","doc":"Checks if a worker is registered.","ref":"Snakepit.Pool.Registry.html#worker_exists?/1"},{"type":"module","title":"Snakepit.Pool.Worker","doc":"GenServer that manages a single external process via Port using adapter pattern.\n\nEach worker:\n- Owns one external process (Python, Node.js, etc.)\n- Handles request/response communication via adapter\n- Manages health checks\n- Reports metrics","ref":"Snakepit.Pool.Worker.html"},{"type":"function","title":"Snakepit.Pool.Worker.busy?/1","doc":"Checks if a worker is busy.","ref":"Snakepit.Pool.Worker.html#busy?/1"},{"type":"function","title":"Snakepit.Pool.Worker.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Snakepit.Pool.Worker.html#child_spec/1"},{"type":"function","title":"Snakepit.Pool.Worker.execute/4","doc":"Executes a command on the worker.","ref":"Snakepit.Pool.Worker.html#execute/4"},{"type":"function","title":"Snakepit.Pool.Worker.get_stats/1","doc":"Gets worker statistics.","ref":"Snakepit.Pool.Worker.html#get_stats/1"},{"type":"function","title":"Snakepit.Pool.Worker.start_link/1","doc":"Starts a worker process.","ref":"Snakepit.Pool.Worker.html#start_link/1"},{"type":"module","title":"Snakepit.Pool.Worker.Starter","doc":"Supervisor wrapper for individual workers that provides automatic restart capability.\n\nThis module implements the \"Permanent Wrapper\" pattern that allows DynamicSupervisor\nto automatically restart workers when they crash, while keeping the actual worker\nprocess as :transient (so it doesn't restart during coordinated shutdown).","ref":"Snakepit.Pool.Worker.Starter.html"},{"type":"module","title":"Architecture - Snakepit.Pool.Worker.Starter","doc":"```\nDynamicSupervisor (WorkerSupervisor)\n└── Worker.Starter (permanent, one per worker)\n    └── Worker (transient, actual worker process)\n```\n\nWhen a Worker crashes:\n1. Worker.Starter detects the crash via its :one_for_one strategy\n2. Worker.Starter automatically restarts the Worker (because Worker is :permanent in this context)\n3. Pool is notified via :DOWN message but doesn't need to manage restarts\n4. New Worker re-registers itself automatically\n\nThis decouples the Pool from worker replacement logic.","ref":"Snakepit.Pool.Worker.Starter.html#module-architecture"},{"type":"function","title":"Snakepit.Pool.Worker.Starter.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Snakepit.Pool.Worker.Starter.html#child_spec/1"},{"type":"function","title":"Snakepit.Pool.Worker.Starter.start_link/1","doc":"Starts a worker starter supervisor.","ref":"Snakepit.Pool.Worker.Starter.html#start_link/1"},{"type":"function","title":"Parameters - Snakepit.Pool.Worker.Starter.start_link/1","doc":"* `worker_id` - Unique identifier for the worker","ref":"Snakepit.Pool.Worker.Starter.html#start_link/1-parameters"},{"type":"function","title":"Snakepit.Pool.Worker.Starter.via_name/1","doc":"Returns a via tuple for this starter supervisor.","ref":"Snakepit.Pool.Worker.Starter.html#via_name/1"},{"type":"module","title":"Snakepit.Pool.Worker.StarterRegistry","doc":"Registry for worker starter supervisors.\n\nThis registry provides a clean separation between worker processes and\ntheir starter supervisors, making debugging and process tracking easier.\n\nWorker starters are registered with their worker_id as the key, allowing\nfor easy lookup and management of individual starter supervisors.","ref":"Snakepit.Pool.Worker.StarterRegistry.html"},{"type":"function","title":"Snakepit.Pool.Worker.StarterRegistry.child_spec/1","doc":"Returns the child spec for the starter registry.","ref":"Snakepit.Pool.Worker.StarterRegistry.html#child_spec/1"},{"type":"function","title":"Snakepit.Pool.Worker.StarterRegistry.get_starter_pid/1","doc":"Gets the PID for a worker starter supervisor.","ref":"Snakepit.Pool.Worker.StarterRegistry.html#get_starter_pid/1"},{"type":"function","title":"Snakepit.Pool.Worker.StarterRegistry.list_starters/0","doc":"Lists all registered worker starter IDs.","ref":"Snakepit.Pool.Worker.StarterRegistry.html#list_starters/0"},{"type":"function","title":"Snakepit.Pool.Worker.StarterRegistry.starter_count/0","doc":"Counts the number of registered worker starters.","ref":"Snakepit.Pool.Worker.StarterRegistry.html#starter_count/0"},{"type":"function","title":"Snakepit.Pool.Worker.StarterRegistry.starter_exists?/1","doc":"Checks if a worker starter is registered.","ref":"Snakepit.Pool.Worker.StarterRegistry.html#starter_exists?/1"},{"type":"function","title":"Snakepit.Pool.Worker.StarterRegistry.via_tuple/1","doc":"Returns a via tuple for registering/looking up a worker starter supervisor.","ref":"Snakepit.Pool.Worker.StarterRegistry.html#via_tuple/1"},{"type":"function","title":"Examples - Snakepit.Pool.Worker.StarterRegistry.via_tuple/1","doc":"iex> Snakepit.Pool.Worker.StarterRegistry.via_tuple(\"worker_123\")\n    {:via, Registry, {Snakepit.Pool.Worker.StarterRegistry, \"worker_123\"}}","ref":"Snakepit.Pool.Worker.StarterRegistry.html#via_tuple/1-examples"},{"type":"module","title":"Snakepit.Pool.WorkerSupervisor","doc":"DynamicSupervisor for pool worker processes.\n\nThis supervisor manages the lifecycle of workers:\n- Starts workers on demand\n- Handles crashes with automatic restarts\n- Provides clean shutdown of workers","ref":"Snakepit.Pool.WorkerSupervisor.html"},{"type":"function","title":"Snakepit.Pool.WorkerSupervisor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Snakepit.Pool.WorkerSupervisor.html#child_spec/1"},{"type":"function","title":"Snakepit.Pool.WorkerSupervisor.list_workers/0","doc":"Lists all supervised workers.","ref":"Snakepit.Pool.WorkerSupervisor.html#list_workers/0"},{"type":"function","title":"Snakepit.Pool.WorkerSupervisor.restart_worker/1","doc":"Restarts a worker by ID.","ref":"Snakepit.Pool.WorkerSupervisor.html#restart_worker/1"},{"type":"function","title":"Snakepit.Pool.WorkerSupervisor.start_link/1","doc":"Starts the worker supervisor.","ref":"Snakepit.Pool.WorkerSupervisor.html#start_link/1"},{"type":"function","title":"Snakepit.Pool.WorkerSupervisor.start_worker/1","doc":"Starts a new pool worker with the given ID.","ref":"Snakepit.Pool.WorkerSupervisor.html#start_worker/1"},{"type":"function","title":"Examples - Snakepit.Pool.WorkerSupervisor.start_worker/1","doc":"iex> Snakepit.Pool.WorkerSupervisor.start_worker(\"worker_123\")\n    {:ok, #PID<0.123.0>}","ref":"Snakepit.Pool.WorkerSupervisor.html#start_worker/1-examples"},{"type":"function","title":"Snakepit.Pool.WorkerSupervisor.stop_worker/1","doc":"Stops a worker gracefully.","ref":"Snakepit.Pool.WorkerSupervisor.html#stop_worker/1"},{"type":"function","title":"Snakepit.Pool.WorkerSupervisor.worker_count/0","doc":"Returns the count of active workers.","ref":"Snakepit.Pool.WorkerSupervisor.html#worker_count/0"},{"type":"module","title":"Snakepit.SessionHelpers","doc":"Session helpers for ML/DSP workflows with program management.\n\nThis module provides domain-specific session functionality for ML frameworks\nthat use program creation and execution patterns (like DSPy, LangChain, etc.).\n\nFor generic session management, use `Snakepit.execute_in_session/4` directly.","ref":"Snakepit.SessionHelpers.html"},{"type":"function","title":"Snakepit.SessionHelpers.execute_program_command/4","doc":"Executes a command in session context with ML program management.\n\nThis function enhances the arguments with session data and handles\npost-processing like storing program metadata for session continuity.","ref":"Snakepit.SessionHelpers.html#execute_program_command/4"},{"type":"function","title":"Examples - Snakepit.SessionHelpers.execute_program_command/4","doc":"# Create a program and store metadata\n    {:ok, response} = Snakepit.SessionHelpers.execute_program_command(\n      \"my_session\", \n      \"create_program\", \n      %{signature: \"input -> output\"}\n    )\n\n    # Execute a program using stored metadata\n    {:ok, result} = Snakepit.SessionHelpers.execute_program_command(\n      \"my_session\",\n      \"execute_program\", \n      %{program_id: \"123\", input: \"data\"}\n    )","ref":"Snakepit.SessionHelpers.html#execute_program_command/4-examples"},{"type":"module","title":"Snakepit.Utils","doc":"Utility functions for Snakepit.\n\nThis module contains common helper functions used across the Snakepit codebase\nto avoid code duplication and provide consistent behavior.","ref":"Snakepit.Utils.html"},{"type":"function","title":"Snakepit.Utils.stringify_keys/1","doc":"Recursively converts atom keys to string keys in maps and lists.\n\nThis is useful when preparing data for JSON serialization where all keys\nneed to be strings.","ref":"Snakepit.Utils.html#stringify_keys/1"},{"type":"function","title":"Examples - Snakepit.Utils.stringify_keys/1","doc":"iex> Snakepit.Utils.stringify_keys(%{foo: \"bar\", baz: %{nested: \"value\"}})\n    %{\"foo\" => \"bar\", \"baz\" => %{\"nested\" => \"value\"}}\n    \n    iex> Snakepit.Utils.stringify_keys([%{key: \"value\"}, %{another: \"test\"}])\n    [%{\"key\" => \"value\"}, %{\"another\" => \"test\"}]\n    \n    iex> Snakepit.Utils.stringify_keys(\"already_string\")\n    \"already_string\"","ref":"Snakepit.Utils.html#stringify_keys/1-examples"},{"type":"extras","title":"Snakepit 🐍","doc":"# Snakepit 🐍\n\n \n   \n \n\n> A high-performance, generalized process pooler and session manager for external language integrations in Elixir\n\n[![CI](https://github.com/nshkrdotcom/snakepit/actions/workflows/ci.yml/badge.svg)](https://github.com/nshkrdotcom/snakepit/actions/workflows/ci.yml)\n[![Hex Version](https://img.shields.io/hexpm/v/snakepit.svg)](https://hex.pm/packages/snakepit)\n[![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)","ref":"readme.html"},{"type":"extras","title":"🚀 What is Snakepit? - Snakepit 🐍","doc":"Snakepit is a battle-tested Elixir library that provides a robust pooling system for managing external processes (Python, Node.js, Ruby, R, etc.). Born from the need for reliable ML/AI integrations, it offers:\n\n- **Lightning-fast concurrent initialization** - 1000x faster than sequential approaches\n- **Session-based execution** with automatic worker affinity\n- **Adapter pattern** for any external language/runtime\n- **Built on OTP primitives** - DynamicSupervisor, Registry, GenServer\n- **Production-ready** with telemetry, health checks, and graceful shutdowns","ref":"readme.html#what-is-snakepit"},{"type":"extras","title":"📋 Table of Contents - Snakepit 🐍","doc":"- [Quick Start](#-quick-start)\n- [Installation](#-installation)\n- [Core Concepts](#-core-concepts)\n- [Configuration](#-configuration)\n- [Usage Examples](#-usage-examples)\n- [Built-in Adapters](#-built-in-adapters)\n- [Creating Custom Adapters](#-creating-custom-adapters)\n- [Session Management](#-session-management)\n- [Monitoring & Telemetry](#-monitoring--telemetry)\n- [Architecture Deep Dive](#-architecture-deep-dive)\n- [Performance](#-performance)\n- [Troubleshooting](#-troubleshooting)\n- [Contributing](#-contributing)","ref":"readme.html#table-of-contents"},{"type":"extras","title":"🏃 Quick Start - Snakepit 🐍","doc":"```elixir\n# In your mix.exs\ndef deps do\n  [\n    {:snakepit, \"~> 0.1.2\"}\n  ]\nend\n\n# Configure and start\nApplication.put_env(:snakepit, :pooling_enabled, true)\nApplication.put_env(:snakepit, :adapter_module, Snakepit.Adapters.GenericPython)\nApplication.put_env(:snakepit, :pool_config, %{pool_size: 4})\n\n{:ok, _} = Application.ensure_all_started(:snakepit)\n\n# Execute commands\n{:ok, result} = Snakepit.execute(\"ping\", %{test: true})\n{:ok, result} = Snakepit.execute(\"compute\", %{operation: \"add\", a: 5, b: 3})\n\n# Session-based execution (maintains state)\n{:ok, result} = Snakepit.execute_in_session(\"user_123\", \"echo\", %{message: \"hello\"})\n```","ref":"readme.html#quick-start"},{"type":"extras","title":"📦 Installation - Snakepit 🐍","doc":"","ref":"readme.html#installation"},{"type":"extras","title":"Hex Package - Snakepit 🐍","doc":"```elixir\ndef deps do\n  [\n    {:snakepit, \"~> 0.1.2\"}\n  ]\nend\n```","ref":"readme.html#hex-package"},{"type":"extras","title":"GitHub (Latest) - Snakepit 🐍","doc":"```elixir\ndef deps do\n  [\n    {:snakepit, github: \"nshkrdotcom/snakepit\"}\n  ]\nend\n```","ref":"readme.html#github-latest"},{"type":"extras","title":"Requirements - Snakepit 🐍","doc":"- Elixir 1.18+\n- Erlang/OTP 27+\n- External runtime (Python 3.8+, Node.js 16+, etc.) depending on adapter","ref":"readme.html#requirements"},{"type":"extras","title":"🎯 Core Concepts - Snakepit 🐍","doc":"","ref":"readme.html#core-concepts"},{"type":"extras","title":"1. Adapters - Snakepit 🐍","doc":"Adapters define how Snakepit communicates with external processes. They specify:\n- The runtime executable (python3, node, ruby, etc.)\n- The bridge script to execute\n- Supported commands and validation\n- Request/response transformations","ref":"readme.html#1-adapters"},{"type":"extras","title":"2. Workers - Snakepit 🐍","doc":"Each worker is a GenServer that:\n- Owns one external process via Erlang Port\n- Handles request/response communication\n- Manages health checks and metrics\n- Auto-restarts on crashes","ref":"readme.html#2-workers"},{"type":"extras","title":"3. Pool - Snakepit 🐍","doc":"The pool manager:\n- Starts workers concurrently on initialization\n- Routes requests to available workers\n- Handles queueing when all workers are busy\n- Supports session affinity for stateful operations","ref":"readme.html#3-pool"},{"type":"extras","title":"4. Sessions - Snakepit 🐍","doc":"Sessions provide:\n- State persistence across requests\n- Worker affinity (same session prefers same worker)\n- TTL-based expiration\n- Centralized storage in ETS","ref":"readme.html#4-sessions"},{"type":"extras","title":"⚙️ Configuration - Snakepit 🐍","doc":"","ref":"readme.html#configuration"},{"type":"extras","title":"Basic Configuration - Snakepit 🐍","doc":"```elixir\n# config/config.exs\nconfig :snakepit,\n  pooling_enabled: true,\n  adapter_module: Snakepit.Adapters.GenericPython,\n  pool_config: %{\n    pool_size: 8  # Default: System.schedulers_online() * 2\n  }\n```","ref":"readme.html#basic-configuration"},{"type":"extras","title":"Advanced Configuration - Snakepit 🐍","doc":"```elixir\nconfig :snakepit,\n  # Pool settings\n  pooling_enabled: true,\n  pool_config: %{\n    pool_size: 16\n  },\n  \n  # Adapter\n  adapter_module: MyApp.CustomAdapter,\n  \n  # Timeouts (milliseconds)\n  pool_startup_timeout: 10_000,      # Max time for worker initialization\n  pool_queue_timeout: 5_000,         # Max time in request queue\n  worker_init_timeout: 20_000,       # Max time for worker to respond to init\n  worker_health_check_interval: 30_000,  # Health check frequency\n  worker_shutdown_grace_period: 2_000,   # Grace period for shutdown\n  \n  # Cleanup settings\n  cleanup_retry_interval: 100,       # Retry interval for cleanup\n  cleanup_max_retries: 10,          # Max cleanup retries\n  \n  # Queue management\n  pool_max_queue_size: 1000         # Max queued requests before rejection\n```","ref":"readme.html#advanced-configuration"},{"type":"extras","title":"Runtime Configuration - Snakepit 🐍","doc":"```elixir\n# Override configuration at runtime\nApplication.put_env(:snakepit, :adapter_module, Snakepit.Adapters.GenericJavaScript)\nApplication.stop(:snakepit)\nApplication.start(:snakepit)\n```","ref":"readme.html#runtime-configuration"},{"type":"extras","title":"📖 Usage Examples - Snakepit 🐍","doc":"","ref":"readme.html#usage-examples"},{"type":"extras","title":"Basic Stateless Execution - Snakepit 🐍","doc":"```elixir\n# Simple computation\n{:ok, %{\"result\" => 8}} = Snakepit.execute(\"compute\", %{\n  operation: \"add\",\n  a: 5,\n  b: 3\n})\n\n# With timeout\n{:ok, result} = Snakepit.execute(\"long_running_task\", %{data: \"...\"}, timeout: 60_000)\n\n# Error handling\ncase Snakepit.execute(\"risky_operation\", %{}) do\n  {:ok, result} -> handle_success(result)\n  {:error, :worker_timeout} -> handle_timeout()\n  {:error, :pool_saturated} -> handle_overload()\n  {:error, reason} -> handle_error(reason)\nend\n```","ref":"readme.html#basic-stateless-execution"},{"type":"extras","title":"Session-Based Execution - Snakepit 🐍","doc":"```elixir\n# Create a session and maintain state\nsession_id = \"user_#{user.id}\"\n\n# First request - initializes session\n{:ok, _} = Snakepit.execute_in_session(session_id, \"initialize\", %{\n  user_id: user.id,\n  preferences: user.preferences\n})\n\n# Subsequent requests use same worker when possible\n{:ok, recommendations} = Snakepit.execute_in_session(session_id, \"get_recommendations\", %{\n  category: \"books\"\n})\n\n# Session data persists across requests\n{:ok, history} = Snakepit.execute_in_session(session_id, \"get_history\", %{})\n```","ref":"readme.html#session-based-execution"},{"type":"extras","title":"ML/AI Workflow Example - Snakepit 🐍","doc":"```elixir\n# Using SessionHelpers for ML program management\nalias Snakepit.SessionHelpers\n\n# Create an ML program/model\n{:ok, response} = SessionHelpers.execute_program_command(\n  \"ml_session_123\",\n  \"create_program\",\n  %{\n    signature: \"question -> answer\",\n    model: \"gpt-3.5-turbo\",\n    temperature: 0.7\n  }\n)\n\nprogram_id = response[\"program_id\"]\n\n# Execute the program multiple times\n{:ok, result} = SessionHelpers.execute_program_command(\n  \"ml_session_123\", \n  \"execute_program\",\n  %{\n    program_id: program_id,\n    input: %{question: \"What is the capital of France?\"}\n  }\n)\n```","ref":"readme.html#ml-ai-workflow-example"},{"type":"extras","title":"Parallel Processing - Snakepit 🐍","doc":"```elixir\n# Process multiple items in parallel across the pool\nitems = [\"item1\", \"item2\", \"item3\", \"item4\", \"item5\"]\n\ntasks = Enum.map(items, fn item ->\n  Task.async(fn ->\n    Snakepit.execute(\"process_item\", %{item: item})\n  end)\nend)\n\nresults = Task.await_many(tasks, 30_000)\n```","ref":"readme.html#parallel-processing"},{"type":"extras","title":"🔌 Built-in Adapters - Snakepit 🐍","doc":"","ref":"readme.html#built-in-adapters"},{"type":"extras","title":"Python Adapter - Snakepit 🐍","doc":"```elixir\n# Configure\nApplication.put_env(:snakepit, :adapter_module, Snakepit.Adapters.GenericPython)\n\n# Available commands\n{:ok, _} = Snakepit.execute(\"ping\", %{})\n{:ok, _} = Snakepit.execute(\"echo\", %{message: \"hello\"})\n{:ok, _} = Snakepit.execute(\"compute\", %{operation: \"multiply\", a: 10, b: 5})\n{:ok, _} = Snakepit.execute(\"info\", %{})\n```","ref":"readme.html#python-adapter"},{"type":"extras","title":"JavaScript/Node.js Adapter - Snakepit 🐍","doc":"```elixir\n# Configure\nApplication.put_env(:snakepit, :adapter_module, Snakepit.Adapters.GenericJavaScript)\n\n# Additional commands\n{:ok, _} = Snakepit.execute(\"random\", %{type: \"uniform\", min: 0, max: 100})\n{:ok, _} = Snakepit.execute(\"compute\", %{operation: \"sqrt\", a: 16})\n```","ref":"readme.html#javascript-node-js-adapter"},{"type":"extras","title":"🛠️ Creating Custom Adapters - Snakepit 🐍","doc":"","ref":"readme.html#creating-custom-adapters"},{"type":"extras","title":"Elixir Adapter Implementation - Snakepit 🐍","doc":"```elixir\ndefmodule MyApp.RubyAdapter do\n  @behaviour Snakepit.Adapter\n  \n  @impl true\n  def executable_path do\n    System.find_executable(\"ruby\")\n  end\n  \n  @impl true\n  def script_path do\n    Path.join(:code.priv_dir(:my_app), \"ruby/bridge.rb\")\n  end\n  \n  @impl true\n  def script_args do\n    [\"--mode\", \"pool-worker\"]\n  end\n  \n  @impl true\n  def supported_commands do\n    [\"ping\", \"process_data\", \"generate_report\"]\n  end\n  \n  @impl true\n  def validate_command(\"process_data\", args) do\n    if Map.has_key?(args, :data) do\n      :ok\n    else\n      {:error, \"Missing required field: data\"}\n    end\n  end\n  \n  def validate_command(\"ping\", _args), do: :ok\n  def validate_command(cmd, _args), do: {:error, \"Unsupported command: #{cmd}\"}\n  \n  # Optional callbacks\n  @impl true\n  def prepare_args(\"process_data\", args) do\n    # Transform args before sending\n    Map.update(args, :data, \"\", &String.trim/1)\n  end\n  \n  @impl true\n  def process_response(\"generate_report\", %{\"report\" => report} = response) do\n    # Post-process the response\n    {:ok, Map.put(response, \"processed_at\", DateTime.utc_now())}\n  end\n  \n  @impl true\n  def command_timeout(\"generate_report\", _args), do: 120_000  # 2 minutes\n  def command_timeout(_command, _args), do: 30_000  # Default 30 seconds\nend\n```","ref":"readme.html#elixir-adapter-implementation"},{"type":"extras","title":"External Bridge Script (Ruby Example) - Snakepit 🐍","doc":"```ruby\n#!/usr/bin/env ruby\n# priv/ruby/bridge.rb\n\nrequire 'json'\n\nclass BridgeHandler\n  def initialize\n    @commands = {\n      'ping' => method(:handle_ping),\n      'process_data' => method(:handle_process_data),\n      'generate_report' => method(:handle_generate_report)\n    }\n  end\n  \n  def run\n    STDERR.puts \"Ruby bridge started\"\n    \n    loop do\n      # Read 4-byte length header\n      length_bytes = STDIN.read(4)\n      break unless length_bytes\n      \n      # Unpack length (big-endian)\n      length = length_bytes.unpack('N')[0]\n      \n      # Read JSON payload\n      json_data = STDIN.read(length)\n      request = JSON.parse(json_data)\n      \n      # Process command\n      response = process_command(request)\n      \n      # Send response\n      json_response = JSON.generate(response)\n      length_header = [json_response.bytesize].pack('N')\n      \n      STDOUT.write(length_header)\n      STDOUT.write(json_response)\n      STDOUT.flush\n    end\n  end\n  \n  private\n  \n  def process_command(request)\n    command = request['command']\n    args = request['args'] || {}\n    \n    handler = @commands[command]\n    if handler\n      result = handler.call(args)\n      {\n        'id' => request['id'],\n        'success' => true,\n        'result' => result,\n        'timestamp' => Time.now.iso8601\n      }\n    else\n      {\n        'id' => request['id'],\n        'success' => false,\n        'error' => \"Unknown command: #{command}\",\n        'timestamp' => Time.now.iso8601\n      }\n    end\n  rescue => e\n    {\n      'id' => request['id'],\n      'success' => false,\n      'error' => e.message,\n      'timestamp' => Time.now.iso8601\n    }\n  end\n  \n  def handle_ping(args)\n    { 'status' => 'ok', 'message' => 'pong' }\n  end\n  \n  def handle_process_data(args)\n    data = args['data'] || ''\n    { 'processed' => data.upcase, 'length' => data.length }\n  end\n  \n  def handle_generate_report(args)\n    # Simulate report generation\n    sleep(1)\n    { \n      'report' => {\n        'title' => args['title'] || 'Report',\n        'generated_at' => Time.now.iso8601,\n        'data' => args['data'] || {}\n      }\n    }\n  end\nend\n\n# Handle signals gracefully\nSignal.trap('TERM') { exit(0) }\nSignal.trap('INT') { exit(0) }\n\n# Run the bridge\nBridgeHandler.new.run\n```","ref":"readme.html#external-bridge-script-ruby-example"},{"type":"extras","title":"🗃️ Session Management - Snakepit 🐍","doc":"","ref":"readme.html#session-management"},{"type":"extras","title":"Session Store API - Snakepit 🐍","doc":"```elixir\nalias Snakepit.Bridge.SessionStore\n\n# Create a session\n{:ok, session} = SessionStore.create_session(\"session_123\", ttl: 7200)\n\n# Store data in session\n:ok = SessionStore.store_program(\"session_123\", \"prog_1\", %{\n  model: \"gpt-4\",\n  temperature: 0.8\n})\n\n# Retrieve session data\n{:ok, session} = SessionStore.get_session(\"session_123\")\n{:ok, program} = SessionStore.get_program(\"session_123\", \"prog_1\")\n\n# Update session\n{:ok, updated} = SessionStore.update_session(\"session_123\", fn session ->\n  Map.put(session, :last_activity, DateTime.utc_now())\nend)\n\n# Check if session exists\ntrue = SessionStore.session_exists?(\"session_123\")\n\n# List all sessions\nsession_ids = SessionStore.list_sessions()\n\n# Manual cleanup\nSessionStore.delete_session(\"session_123\")\n\n# Get session statistics\nstats = SessionStore.get_stats()\n```","ref":"readme.html#session-store-api"},{"type":"extras","title":"Global Program Storage - Snakepit 🐍","doc":"```elixir\n# Store programs accessible by any worker\n:ok = SessionStore.store_global_program(\"template_1\", %{\n  type: \"qa_template\",\n  prompt: \"Answer the following question: {question}\"\n})\n\n# Retrieve from any worker\n{:ok, template} = SessionStore.get_global_program(\"template_1\")\n```","ref":"readme.html#global-program-storage"},{"type":"extras","title":"📊 Monitoring & Telemetry - Snakepit 🐍","doc":"","ref":"readme.html#monitoring-telemetry"},{"type":"extras","title":"Available Events - Snakepit 🐍","doc":"```elixir\n# Worker request completed\n[:snakepit, :worker, :request]\n# Measurements: %{duration: milliseconds}\n# Metadata: %{result: :ok | :error}\n\n# Worker initialized\n[:snakepit, :worker, :initialized]\n# Measurements: %{initialization_time: seconds}\n# Metadata: %{worker_id: string}\n```","ref":"readme.html#available-events"},{"type":"extras","title":"Setting Up Monitoring - Snakepit 🐍","doc":"```elixir\n# In your application startup\n:telemetry.attach_many(\n  \"snakepit-metrics\",\n  [\n    [:snakepit, :worker, :request],\n    [:snakepit, :worker, :initialized]\n  ],\n  &MyApp.Metrics.handle_event/4,\n  %{}\n)\n\ndefmodule MyApp.Metrics do\n  require Logger\n  \n  def handle_event([:snakepit, :worker, :request], measurements, metadata, _config) do\n    # Log slow requests\n    if measurements.duration > 5000 do\n      Logger.warning(\"Slow request: #{measurements.duration}ms\")\n    end\n    \n    # Send to StatsD/Prometheus/DataDog\n    MyApp.Metrics.Client.histogram(\n      \"snakepit.request.duration\",\n      measurements.duration,\n      tags: [\"result:#{metadata.result}\"]\n    )\n  end\n  \n  def handle_event([:snakepit, :worker, :initialized], measurements, metadata, _config) do\n    Logger.info(\"Worker #{metadata.worker_id} started in #{measurements.initialization_time}s\")\n  end\nend\n```","ref":"readme.html#setting-up-monitoring"},{"type":"extras","title":"Pool Statistics - Snakepit 🐍","doc":"```elixir\nstats = Snakepit.get_stats()\n# Returns:\n# %{\n#   workers: 8,          # Total workers\n#   available: 6,        # Available workers\n#   busy: 2,            # Busy workers\n#   requests: 1534,     # Total requests\n#   queued: 0,          # Currently queued\n#   errors: 12,         # Total errors\n#   queue_timeouts: 3,  # Queue timeout count\n#   pool_saturated: 0   # Saturation rejections\n# }\n```","ref":"readme.html#pool-statistics"},{"type":"extras","title":"🏗️ Architecture Deep Dive - Snakepit 🐍","doc":"","ref":"readme.html#architecture-deep-dive"},{"type":"extras","title":"Component Overview - Snakepit 🐍","doc":"```\n┌───────────────────────────────────────────────────────┐\n│                    Snakepit Application               │\n├───────────────────────────────────────────────────────┤\n│                                                       │\n│  ┌─────────────┐  ┌──────────────┐  ┌───────────────┐ │\n│  │    Pool     │  │ SessionStore │  │  Registries   │ │\n│  │  Manager    │  │   (ETS)      │  │  (Worker/Proc)│ │\n│  └──────┬──────┘  └──────────────┘  └───────────────┘ │\n│         │                                             │\n│  ┌──────▼────────────────────────────────────────────┐│\n│  │            WorkerSupervisor (Dynamic)             ││\n│  └──────┬────────────────────────────────────────────┘│\n│         │                                             │\n│  ┌──────▼──────┐  ┌──────────────┐  ┌──────────────┐  │\n│  │   Worker    │  │   Worker     │  │   Worker     │  │\n│  │  Starter    │  │  Starter     │  │  Starter     │  │\n│  │(Supervisor) │  │(Supervisor)  │  │(Supervisor)  │  │\n│  └──────┬──────┘  └───────┬──────┘  └───────┬──────┘  │\n│         │                 │                 │         │\n│  ┌──────▼──────┐  ┌───────▼──────┐  ┌───────▼──────┐  │\n│  │   Worker    │  │   Worker     │  │   Worker     │  │\n│  │ (GenServer) │  │ (GenServer)  │  │ (GenServer)  │  │\n│  └──────┬──────┘  └───────┬──────┘  └───────┬──────┘  │\n│         │                 │                 │         │\n└─────────┼─────────────────┼─────────────────┼─────────┘\n          │                 │                 │\n    ┌─────▼──────┐    ┌─────▼──────┐    ┌─────▼──────┐\n    │  External  │    │  External  │    │  External  │\n    │  Process   │    │  Process   │    │  Process   │\n    │  (Python)  │    │  (Node.js) │    │   (Ruby)   │\n    └────────────┘    └────────────┘    └────────────┘\n```","ref":"readme.html#component-overview"},{"type":"extras","title":"Key Design Decisions - Snakepit 🐍","doc":"1. **Concurrent Initialization**: Workers start in parallel using `Task.async_stream`\n2. **Permanent Wrapper Pattern**: Worker.Starter supervises Workers for auto-restart\n3. **Centralized State**: All session data in ETS, workers are stateless\n4. **Registry-Based**: O(1) worker lookups and reverse PID lookups\n5. **Port Communication**: Binary protocol with 4-byte length headers","ref":"readme.html#key-design-decisions"},{"type":"extras","title":"Process Lifecycle - Snakepit 🐍","doc":"1. **Startup**:\n   - Pool manager starts\n   - Concurrently spawns N workers via WorkerSupervisor\n   - Each worker starts its external process\n   - Workers send init ping and register when ready\n\n2. **Request Flow**:\n   - Client calls `Snakepit.execute/3`\n   - Pool finds available worker (with session affinity if applicable)\n   - Worker sends request to external process\n   - External process responds\n   - Worker returns result to client\n\n3. **Crash Recovery**:\n   - Worker crashes → Worker.Starter restarts it automatically\n   - External process dies → Worker detects and crashes → restart\n   - Pool crashes → Supervisor restarts entire pool\n\n4. **Shutdown**:\n   - Pool manager sends shutdown to all workers\n   - Workers close ports gracefully (SIGTERM)\n   - ApplicationCleanup ensures no orphaned processes (SIGKILL)","ref":"readme.html#process-lifecycle"},{"type":"extras","title":"⚡ Performance - Snakepit 🐍","doc":"","ref":"readme.html#performance"},{"type":"extras","title":"Benchmarks - Snakepit 🐍","doc":"```\nConfiguration: 16 workers, Python adapter\nHardware: 8-core CPU, 32GB RAM\n\nStartup Time:\n- Sequential: 16 seconds (1s per worker)\n- Concurrent: 1.2 seconds (13x faster)\n\nThroughput:\n- Simple computation: 50,000 req/s\n- Complex ML inference: 1,000 req/s\n- Session operations: 45,000 req/s\n\nLatency (p99):\n- Simple computation: < 2ms\n- Complex ML inference: < 100ms\n- Session operations: < 1ms\n```","ref":"readme.html#benchmarks"},{"type":"extras","title":"Optimization Tips - Snakepit 🐍","doc":"1. **Pool Size**: Start with `System.schedulers_online() * 2`\n2. **Queue Size**: Monitor `pool_saturated` errors and adjust\n3. **Timeouts**: Set appropriate timeouts per command type\n4. **Session TTL**: Balance memory usage vs cache hits\n5. **Health Checks**: Increase interval for stable workloads","ref":"readme.html#optimization-tips"},{"type":"extras","title":"🔧 Troubleshooting - Snakepit 🐍","doc":"","ref":"readme.html#troubleshooting"},{"type":"extras","title":"Common Issues - Snakepit 🐍","doc":"#### Workers Not Starting\n\n```elixir\n# Check adapter configuration\nadapter = Application.get_env(:snakepit, :adapter_module)\nadapter.executable_path()  # Should return valid path\nFile.exists?(adapter.script_path())  # Should return true\n\n# Check logs for errors\nLogger.configure(level: :debug)\n```\n\n#### Port Exits\n\n```elixir\n# Enable port tracing\n:erlang.trace(Process.whereis(Snakepit.Pool.Worker), true, [:receive, :send])\n\n# Check external process logs\n# Python: Add logging to bridge script\n# Node.js: Check stderr output\n```\n\n#### Memory Leaks\n\n```elixir\n# Monitor ETS usage\n:ets.info(:snakepit_sessions, :memory)\n\n# Check for orphaned processes\nSnakepit.Pool.ProcessRegistry.get_stats()\n\n# Force cleanup\nSnakepit.Bridge.SessionStore.cleanup_expired_sessions()\n```","ref":"readme.html#common-issues"},{"type":"extras","title":"Debug Mode - Snakepit 🐍","doc":"```elixir\n# Enable debug logging\nLogger.configure(level: :debug)\n\n# Trace specific worker\n:sys.trace(Snakepit.Pool.Registry.via_tuple(\"worker_1\"), true)\n\n# Get internal state\n:sys.get_state(Snakepit.Pool)\n```","ref":"readme.html#debug-mode"},{"type":"extras","title":"🤝 Contributing - Snakepit 🐍","doc":"We welcome contributions! Please see our [Contributing Guide](https://github.com/nshkrdotcom/snakepit/blob/main/CONTRIBUTING.md) for details.","ref":"readme.html#contributing"},{"type":"extras","title":"Development Setup - Snakepit 🐍","doc":"```bash\n# Clone the repo\ngit clone https://github.com/nshkrdotcom/snakepit.git\ncd snakepit\n\n# Install dependencies\nmix deps.get\n\n# Run tests\nmix test\n\n# Run with example scripts\nelixir examples/session_based_demo.exs\nelixir examples/javascript_session_demo.exs\n\n# Check code quality\nmix format --check-formatted\nmix dialyzer\n```","ref":"readme.html#development-setup"},{"type":"extras","title":"Running Tests - Snakepit 🐍","doc":"```bash\n# All tests\nmix test\n\n# With coverage\nmix test --cover\n\n# Specific test\nmix test test/snakepit_test.exs:42\n```","ref":"readme.html#running-tests"},{"type":"extras","title":"📝 License - Snakepit 🐍","doc":"Snakepit is released under the MIT License. See the [LICENSE](https://github.com/nshkrdotcom/snakepit/blob/main/LICENSE) file for details.","ref":"readme.html#license"},{"type":"extras","title":"🙏 Acknowledgments - Snakepit 🐍","doc":"- Inspired by the need for reliable ML/AI integrations in Elixir\n- Built on battle-tested OTP principles\n- Special thanks to the Elixir community","ref":"readme.html#acknowledgments"},{"type":"extras","title":"📚 Resources - Snakepit 🐍","doc":"- [Hex Package](https://hex.pm/packages/snakepit)\n- [API Documentation](https://hexdocs.pm/snakepit)\n- [GitHub Repository](https://github.com/nshkrdotcom/snakepit)\n- [Example Projects](https://github.com/nshkrdotcom/snakepit/tree/main/examples)\n\n---\n\nMade with ❤️ by [NSHkr](https://github.com/nshkrdotcom)","ref":"readme.html#resources"},{"type":"extras","title":"DIAGS","doc":"","ref":"diags.html"},{"type":"extras","title":"1. Overall System Architecture - High Performance Overview - DIAGS","doc":"```mermaid\ngraph LR\n    ElixirOTPLayer[\"Elixir OTP Layer\"]\n    subgraph ElixirOTPLayer\n        Client[\"Client API\"]\n        Pool[\"Pool Manager ⚡ Non-blocking async\"]\n        TaskSup[\"Task Supervisor ⚡ Isolated execution\"]\n        \n        WorkerManagement[\"Worker Management\"]\n        subgraph WorkerManagement\n            WorkerSup[\"Worker Supervisor ⚡ Dynamic workers\"]\n            Starter1[\"Worker Starter 1 🔄 Auto-restart\"]\n            Starter2[\"Worker Starter 2 🔄 Auto-restart\"]\n            StarterN[\"Worker Starter N 🔄 Auto-restart\"]\n            Worker1[\"Worker 1 GenServer\"]\n            Worker2[\"Worker 2 GenServer\"]\n            WorkerN[\"Worker N GenServer\"]\n        end\n        \n        HighPerformanceRegistries[\"High Performance Registries\"]\n        subgraph HighPerformanceRegistries\n            Registry[\"Worker Registry ⚡ O(1) lookups\"]\n            ProcReg[\"Process Registry ⚡ PID tracking\"]\n            StarterReg[\"Starter Registry ⚡ Supervisor tracking\"]\n        end\n        \n        SessionStoreETS[\"Session Store ETS\"]\n        subgraph SessionStoreETS\n            SessionStore[\"Session Store ⚡ Concurrent R/W 📊 Decentralized counters\"]\n            GlobalPrograms[\"Global Programs ⚡ Public table access\"]\n        end\n    end\n    \n    ExternalProcesses[\"External Processes\"]\n    subgraph ExternalProcesses\n        Python1[\"Python Process 1 🐍 Port communication\"]\n        Python2[\"Python Process 2 🐍 Port communication\"]\n        PythonN[\"Python Process N 🐍 Port communication\"]\n    end\n    \n    Client -->|\"⚡ Async call\"| Pool\n    Pool -->|\"⚡ O(1) checkout\"| Registry\n    Pool -->|\"🎯 Session affinity\"| SessionStore\n    Pool -->|\"⚡ Task.async_nolink\"| TaskSup\n    TaskSup -->|Execute| Worker1\n    TaskSup -->|Execute| Worker2\n    TaskSup -->|Execute| WorkerN\n    \n    WorkerSup -->|Supervise| Starter1\n    WorkerSup -->|Supervise| Starter2\n    WorkerSup -->|Supervise| StarterN\n    \n    Starter1 -->|Auto-restart| Worker1\n    Starter2 -->|Auto-restart| Worker2\n    StarterN -->|Auto-restart| WorkerN\n    \n    Worker1 -->|Binary protocol 4-byte frames| Python1\n    Worker2 -->|Binary protocol 4-byte frames| Python2\n    WorkerN -->|Binary protocol 4-byte frames| PythonN\n    \n    Worker1 -->|Register| Registry\n    Worker2 -->|Register| Registry\n    WorkerN -->|Register| Registry\n    \n    Worker1 -->|Track PID| ProcReg\n    Worker2 -->|Track PID| ProcReg\n    WorkerN -->|Track PID| ProcReg\n    \n    style ElixirOTPLayer fill:#e6e6fa,stroke:#9370db,stroke-width:3px,color:#2e1065\n    style WorkerManagement fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style HighPerformanceRegistries fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style SessionStoreETS fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style ExternalProcesses fill:#f0f9ff,stroke:#0ea5e9,stroke-width:2px,color:#0c4a6e\n    style Pool fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style SessionStore fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style Registry fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style TaskSup fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n```","ref":"diags.html#1-overall-system-architecture-high-performance-overview"},{"type":"extras","title":"2. Request Flow - Performance Critical Path - DIAGS","doc":"```mermaid\nsequenceDiagram\n    participant C as Client\n    participant P as Pool ⚡ Non-blocking\n    participant TS as TaskSupervisor ⚡ Isolated\n    participant R as Registry ⚡ O(1)\n    participant S as SessionStore ⚡ ETS\n    participant W as Worker\n    participant E as External Process\n    \n    C->>P: execute(command, args)\n    \n    alt Session-based request\n        P->>S: get_preferred_worker ⚡ O(1) ETS lookup\n        S-->>P: worker_id or nil\n    end\n    \n    P->>R: checkout_worker ⚡ O(1) via Registry\n    R-->>P: worker_id\n    \n    P->>TS: Task.async_nolink ⚡ Non-blocking\n    Note over P: Pool returns immediately to handle next request\n    \n    TS->>W: GenServer.call\n    W->>E: Port.command ⚡ Binary protocol\n    E-->>W: Response ⚡ 4-byte framed\n    W-->>TS: Result\n    TS-->>C: GenServer.reply ⚡ Direct to client\n    \n    TS->>P: checkin_worker ⚡ Cast (async)\n    \n    alt Queued requests exist\n        P->>P: Process next from queue\n    else No queued requests\n        P->>R: Mark available ⚡ O(1) update\n    end\n```","ref":"diags.html#2-request-flow-performance-critical-path"},{"type":"extras","title":"3. ETS Tables Architecture - High Performance Storage - DIAGS","doc":"```mermaid\ngraph LR\n    SessionStoreETSTables[\"Session Store ETS Tables\"]\n    subgraph SessionStoreETSTables\n        SessionsTable[\"Sessions Table\"]\n        subgraph SessionsTable\n            ST[\":snakepit_sessions ⚡ read_concurrency: true ⚡ write_concurrency: true ⚡ decentralized_counters: true\"]\n            \n            S1[\"Key: session_1 Value: {last_accessed, ttl, session_data}\"]\n            S2[\"Key: session_2 Value: {last_accessed, ttl, session_data}\"]\n            SN[\"Key: session_N Value: {last_accessed, ttl, session_data}\"]\n        end\n        \n        GlobalProgramsTable[\"Global Programs Table\"]\n        subgraph GlobalProgramsTable\n            GP[\":snakepit_sessions_global_programs ⚡ Same optimizations\"]\n            \n            P1[\"Key: program_1 Value: {data, timestamp}\"]\n            P2[\"Key: program_2 Value: {data, timestamp}\"]\n            PN[\"Key: program_N Value: {data, timestamp}\"]\n        end\n    end\n    \n    OptimizedOperations[\"Optimized Operations\"]\n    subgraph OptimizedOperations\n        Read[\"⚡ Concurrent reads No locking\"]\n        Write[\"⚡ Concurrent writes Decentralized counters\"]\n        Cleanup[\"⚡ select_delete Atomic batch cleanup\"]\n    end\n    \n    ST --> S1\n    ST --> S2\n    ST --> SN\n    \n    GP --> P1\n    GP --> P2\n    GP --> PN\n    \n    Read --> ST\n    Read --> GP\n    Write --> ST\n    Write --> GP\n    Cleanup --> ST\n    Cleanup --> GP\n    \n    style SessionStoreETSTables fill:#e6e6fa,stroke:#9370db,stroke-width:3px,color:#2e1065\n    style SessionsTable fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style GlobalProgramsTable fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style OptimizedOperations fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style ST fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style GP fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style Read fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style Write fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style Cleanup fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n```","ref":"diags.html#3-ets-tables-architecture-high-performance-storage"},{"type":"extras","title":"4. Worker Lifecycle - Performance & Reliability - DIAGS","doc":"```mermaid\nstateDiagram-v2\n    [*] --> Starting: Pool requests worker\n    \n    Starting --> Initializing: Port opened ⚡ Parallel startup\n    \n    Initializing --> Ready: Init ping OK 📊 Telemetry emitted\n    Initializing --> Failed: Timeout/Error\n    \n    Ready --> Busy: Request received ⚡ O(1) checkout\n    Busy --> Ready: Response sent ⚡ O(1) checkin\n    \n    Ready --> HealthCheck: Periodic check ⏱️ Every 30s\n    HealthCheck --> Ready: Healthy\n    HealthCheck --> Unhealthy: Failed\n    \n    Unhealthy --> Restarting: Supervisor detects\n    Failed --> Restarting: Auto-restart\n    \n    Restarting --> Starting: ♻️ Via Starter\n    \n    Ready --> Terminating: Shutdown signal\n    Busy --> Terminating: Graceful shutdown\n    \n    Terminating --> [*]: Process cleaned up\n    \n    note right of Ready\n        ⚡ Worker pool maintains\n        hot workers ready for\n        immediate use\n    end note\n    \n    note right of Busy\n        ⚡ Non-blocking async\n        execution via Task\n        Supervisor\n    end note\n```","ref":"diags.html#4-worker-lifecycle-performance-reliability"},{"type":"extras","title":"5. Concurrent Initialization Performance - DIAGS","doc":"```mermaid\ngraph TD\n    SequentialStartupTraditional[\"Sequential Startup Traditional\"]\n    subgraph SequentialStartupTraditional\n        T0[\"Start\"] --> W1S[\"Worker 1 2s\"]\n        W1S --> W2S[\"Worker 2 2s\"]\n        W2S --> W3S[\"Worker 3 2s\"]\n        W3S --> W4S[\"Worker 4 2s\"]\n        W4S --> DoneS[\"Ready Total: 8s\"]\n    end\n    \n    ConcurrentStartupSnakepit[\"Concurrent Startup Snakepit\"]\n    subgraph ConcurrentStartupSnakepit\n        T0C[\"Start\"] --> Init[\"Task.async_stream\"]\n        Init --> W1C[\"Worker 1 2s\"]\n        Init --> W2C[\"Worker 2 2s\"]\n        Init --> W3C[\"Worker 3 2s\"]\n        Init --> W4C[\"Worker 4 2s\"]\n        \n        W1C --> Collect\n        W2C --> Collect\n        W3C --> Collect\n        W4C --> Collect\n        \n        Collect --> DoneC[\"Ready Total: ~2s\"]\n    end\n    \n    style SequentialStartupTraditional fill:#fef3c7,stroke:#f59e0b,stroke-width:2px,color:#92400e\n    style ConcurrentStartupSnakepit fill:#e6e6fa,stroke:#9370db,stroke-width:3px,color:#2e1065\n    style Init fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style DoneC fill:#dcfce7,stroke:#22c55e,stroke-width:2px,color:#14532d\n    style DoneS fill:#fee2e2,stroke:#ef4444,stroke-width:2px,color:#7f1d1d\n```","ref":"diags.html#5-concurrent-initialization-performance"},{"type":"extras","title":"6. Request Queueing & Load Distribution - DIAGS","doc":"```mermaid\ngraph LR\n    HighPerformanceRequestHandling[\"High Performance Request Handling\"]\n    subgraph HighPerformanceRequestHandling\n        RequestQueue[\"Request Queue\"]\n        subgraph RequestQueue\n            Q[\":queue (Erlang) ⚡ FIFO ⚡ O(1) operations\"]\n            R1[\"Request 1\"]\n            R2[\"Request 2\"]\n            R3[\"Request 3\"]\n            RN[\"Request N\"]\n        end\n        \n        WorkerPoolState[\"Worker Pool State\"]\n        subgraph WorkerPoolState\n            Available[\"MapSet ⚡ O(1) member check ⚡ O(1) add/remove\"]\n            Busy[\"Map ⚡ O(1) lookup\"]\n            \n            AW1[\"Worker 1\"]\n            AW2[\"Worker 2\"]\n            BW3[\"Worker 3 🔴\"]\n            BW4[\"Worker 4 🔴\"]\n        end\n        \n        LoadDistribution[\"Load Distribution\"]\n        subgraph LoadDistribution\n            Check{\"Worker Available?\"}\n            Assign[\"Assign to worker ⚡ O(1)\"]\n            Queue[\"Queue request ⚡ O(1)\"]\n            Dequeue[\"Process from queue ⚡ O(1)\"]\n        end\n    end\n    \n    R1 --> Check\n    R2 --> Check\n    R3 --> Check\n    RN --> Check\n    \n    Check -->|Yes| Assign\n    Check -->|No| Queue\n    \n    Queue --> Q\n    Q --> Dequeue\n    \n    Assign --> Available\n    Available --> AW1\n    Available --> AW2\n    \n    Busy --> BW3\n    Busy --> BW4\n    \n    Dequeue -->|Worker freed| Assign\n    \n    style HighPerformanceRequestHandling fill:#e6e6fa,stroke:#9370db,stroke-width:3px,color:#2e1065\n    style RequestQueue fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style WorkerPoolState fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style LoadDistribution fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style Q fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style Available fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style Check fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n```","ref":"diags.html#6-request-queueing-load-distribution"},{"type":"extras","title":"7. Process Registry - O(1) Performance - DIAGS","doc":"```mermaid\ngraph LR\n    RegistryArchitecture[\"Registry Architecture\"]\n    subgraph RegistryArchitecture\n        WorkerRegistry[\"Worker Registry\"]\n        subgraph WorkerRegistry\n            WR[\"Elixir Registry ⚡ :unique keys ⚡ O(1) operations\"]\n            WK1[\"worker_1 → PID1\"]\n            WK2[\"worker_2 → PID2\"]\n            WKN[\"worker_N → PIDN\"]\n        end\n        \n        ProcessRegistryETS[\"Process Registry ETS\"]\n        subgraph ProcessRegistryETS\n            PR[\"Process Registry ⚡ :protected table ⚡ read_concurrency\"]\n            PK1[\"worker_1 → {pid, os_pid, fingerprint}\"]\n            PK2[\"worker_2 → {pid, os_pid, fingerprint}\"]\n            PKN[\"worker_N → {pid, os_pid, fingerprint}\"]\n        end\n        \n        StarterRegistry[\"Starter Registry\"]\n        subgraph StarterRegistry\n            SR[\"Starter Registry ⚡ Supervisor tracking\"]\n            SK1[\"worker_1 → Starter PID1\"]\n            SK2[\"worker_2 → Starter PID2\"]\n            SKN[\"worker_N → Starter PIDN\"]\n        end\n    end\n    \n    O1Operations[\"O(1) Operations\"]\n    subgraph O1Operations\n        Op1[\"via_tuple lookup ⚡ Direct to worker\"]\n        Op2[\"Reverse lookup ⚡ PID to worker_id\"]\n        Op3[\"OS PID tracking ⚡ Cleanup guarantee\"]\n    end\n    \n    WR --> WK1\n    WR --> WK2\n    WR --> WKN\n    \n    PR --> PK1\n    PR --> PK2\n    PR --> PKN\n    \n    SR --> SK1\n    SR --> SK2\n    SR --> SKN\n    \n    Op1 --> WR\n    Op2 --> WR\n    Op3 --> PR\n    \n    style RegistryArchitecture fill:#e6e6fa,stroke:#9370db,stroke-width:3px,color:#2e1065\n    style WorkerRegistry fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style ProcessRegistryETS fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style StarterRegistry fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style O1Operations fill:#f3e8ff,stroke:#a855f7,stroke-width:2px,color:#581c87\n    style WR fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style PR fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n    style SR fill:#ddd6fe,stroke:#7c3aed,stroke-width:2px,color:#581c87\n```","ref":"diags.html#7-process-registry-o-1-performance"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.2"}}